#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include "skyvers.h"
#include "skyskip.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>


/*
*  This program reads data from a serial port and dumps it into a file.
*
*  Copyright (C) 2018 - 2021 R.J. van der Putten, Leiden, Holland, 
*  rob at sput dot nl.
* 
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License version 3 as
*  published by the Free Software Foundation.
* 
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*/

/*  TTY  */
int fd;
struct termios oldtio;

/*  Use timeout or pipe for SIGTERM  */
//#define SKLRD_USE_TIMEOUT 1

#ifndef SKLRD_USE_TIMEOUT
 /*  Pipe for select  */
 int pfd[2];
#endif

struct my_baud {
    unsigned int rate;
    char *str;
} my_bauds[] = { \
    {    B300,    "300" }, \
    {    B600,    "600" }, \
    {   B1200,   "1200" }, \
    {   B2400,   "2400" }, \
    {   B4800,   "4800" }, \
    {   B9600,   "9600" }, \
    {  B19200,  "19200" }, \
    {  B38400,  "38400" }, \
    {  B57600,  "57600" }, \
    { B115200, "115200" }, \
    {       0,    NULL  }  \
};

#define SKLRD_DFLT_SPD "115200"

/*  Global stuff  */
int contprog;	/*  Stop program when 0  */

/*  Date - Time  */
char datstr[48];
struct timeval dstv;

/*  Logging  */
char logstr[512];
char tologst[768];
pid_t pid;
FILE *logfil;

/*  Mail  */
char systr[768];


/*
    inspired by
    https://tldp.org/HOWTO/Serial-Programming-HOWTO/x115.html
*/


/*  Generate date string */
char * gendat(char *s, int mode)
{
    int len, ms;
    struct tm *time;

    if (s == NULL)
	return(NULL);

    /*
    *	Modes;
    *
    *	0  YYYYMMDD-hhmmss
    *	1  YYYY-MM-DD hh:mm:ss
    *	2  YYYYMMDD-hhmmss-ms
    *	3  YYYY-MM-DD hh:mm:ss.ms
    */

    gettimeofday(&dstv, NULL);
    ms = (int) ((long) dstv.tv_usec / 1000L);
    time = localtime(&dstv.tv_sec);

    if ((mode & 1) == 0) {
	len = snprintf(s, 20, "%04d%02d%02d-%02d%02d%02d", \
	  time->tm_year + 1900, time->tm_mon + 1, time->tm_mday, \
	  time->tm_hour, time->tm_min, time->tm_sec);
    } else {
	len = snprintf(s, 20, "%04d-%02d-%02d %02d:%02d:%02d", \
	  time->tm_year + 1900, time->tm_mon + 1, time->tm_mday, \
	  time->tm_hour, time->tm_min, time->tm_sec);
    }

    if ((mode & 2) != 0) {
	/*  Add milliseconds  */
	if ((mode & 1) == 0)
	    snprintf(s + len, 12, "-%03d", ms);
	else
	    snprintf(s + len, 12, ".%03d", ms);
    }

    return(s);
}

/*  Print to logfile or stderr  */
int prtolog(char *s1, char *s2)
{
    if(s1 == NULL)
        return(1);

    if (logfil == NULL) {
	/*  Interactive mode, print to stderr  */
	fprintf(stderr, "%s", s1);
	if (s2 != NULL)
	    fprintf(stderr, " %s", s2);
	fprintf(stderr, "\n");
	fflush(stderr);
    } else {
	/*  Deamon, print to log  */
	if(s2 == NULL)
	    snprintf(tologst, 767, "%s %s\n", gendat(datstr, 3), s1);
	else
	    snprintf(tologst, 767, "%s %s %s\n", gendat(datstr, 3), s1, s2);
	fprintf(logfil, "%s", tologst);
	fflush(logfil);
    }

    return(0);
}

/*  Stop program  */
void stop(int sig)
{
    #ifndef SKLRD_USE_TIMEOUT
     char ch;
     ch = 'q';
    #endif

    contprog = 0;
    prtolog("stop(): recieved SIGTERM", NULL);
    #ifndef SKLRD_USE_TIMEOUT
     /*  Write to pipe to make sure */
     if (write(pfd[1], &ch, 1) < 0) {
	 snprintf(logstr, 511, "stop(): Can't write to pipe: %s", \
	   strerror(errno));
	 prtolog(logstr, NULL);
     }
    #endif
}

/*  Reset TTY  */
void resetty(void)
{
    tcflush(fd, TCIOFLUSH);
    tcsetattr(fd, TCSANOW, &oldtio);
    close(fd);
}

/*  Generate filename and open file  */
FILE * genfil(FILE *f1, char *filstr, char *path, char *prefix)
{
    /* f1 = genfil(f1, filstr, path, name); */

    if(path != NULL && path[0] != 0)
	snprintf(filstr, 383, "%s/%s-%s.txt", path, prefix, gendat(datstr, 2));
    else 
	snprintf(filstr, 383, "%s-%s.txt", prefix, gendat(datstr, 2));

    if (filstr[382] != 0) {
	prtolog("genfil(): Filename or path too long", NULL);
	return(NULL);
    }
    if ((f1 = fopen(filstr, "a")) == NULL) {
	snprintf(logstr, 511, "genfil(): Can't open %s: %s", \
	  filstr, strerror(errno));
	prtolog(logstr, NULL);
    }
 
    return(f1);
}

/*  Help  */
void hlp(void)
{
    fprintf(stderr, "Writes data from Skyla VB1 to file.\n");
    fprintf(stderr, "Usage: skyla-read [-a altpath] [-b baudrate] [-c] ");
    fprintf(stderr, "-d device [-D] [-g gid] [-h]\n");
    fprintf(stderr, "\t[-i] [-l logfile] [-m address] ");
    fprintf(stderr, "[-n filename prexfix] [-p path]\n");
    fprintf(stderr, "\t[-r piddir] [-s] [-u uid] [-v] [-x]\n");
    fprintf(stderr, "-c: Enable RTS CTS flow control.\n");
    fprintf(stderr, "-D: Enable debug.\n");
    fprintf(stderr, "-h: Print help and exit.\n");
    fprintf(stderr, "-i: Ignore stdin.\n");
    fprintf(stderr, "-s: Strip output.\n");
    fprintf(stderr, "-v: Print version and exit.\n");
    fprintf(stderr, "-x: Enable XON XOFF flow control.\n");
    fprintf(stderr, "Default baudrate is %s bps.\n", SKLRD_DFLT_SPD);
    fprintf(stderr, "Default filename is TTY-YYYYMMDD-hhmmss-ms.txt,\n");
    fprintf(stderr, "\tEG: ttyS1-20181224-135557-123.txt\n");
    fprintf(stderr, "The tty is set to RAW, 8N1.\n");
}

/*  Print version  */
void prtvers(void)
{
    printf("Version: %s\n", skyvers);
    #ifdef SKLRD_USE_TIMEOUT
     printf("Compiled with use timeout.\n");
    #endif
}

/* Convert baudrate into int  */
unsigned int findbaud(char *s)
{
    int i;

    i = 0;
    while (my_bauds[i].str != NULL) {
	if(strcmp(s, my_bauds[i].str) == 0)
	    return(my_bauds[i].rate);
	i++;
    }

    return(0);
}

/*  List baudrates  */
void listbauds(void)
{
    int i;

    fprintf(stderr, "%s", my_bauds[0].str);
    i = 1;
    while (my_bauds[i].str != NULL) {
	fprintf(stderr, ", %s", my_bauds[i].str);
	i++;
    }
}

/*  Mail errors  */
int mailerr(char *mes, char *adr)
{
    if(mes == NULL || adr == NULL)
	return(1);

    snprintf(systr, 767, "echo \"%s\" | mail -s \"Skyla error\" \"%s\"", \
	mes, adr);
    system(systr);

    return(0);
}

/* See if line needs to be skipped  */
int fndskip(char *s)
{
    int i;

    if(s == NULL)
	return(0);

    i = 0;
    while (skiplist[i].len != 0) {
	if (strncasecmp(s, skiplist[i].str, skiplist[i].len) == 0) {
	    /*  Found in skiplist  */
	    return(1);
	}
	i++;
    }

    return(0);
}

/*  Strip trailing whitespace  */
int strptws(char *s, int len)
{
    int i;

    if(s == NULL || s[0] == 0)
	return(0);
    if(s[0] <= ' ' && s[0] > 0)
	return(len);

    i = len;
    while (i >= 0) {
	if (s[i] == 13 && s[i + 1] == 10) {
	    /*  Convert CrLf to Lf  */
	    s[i] = 10;
	    s[i + 1] = 0;
	} else if (s[i] > ' ' || s[i] < 0) {
	    /*  No whitespace  */
	    break;
	}
	i--;
    }
    /*  Terminate  */
    s[i + 1] = 10;
    s[i + 2] = 0;

    return(i + 1);
}


int main(int argc, char **argv)
{
    int ch, i, len, maxfd, op, opt, retcnt, retval, tot;
    int debfl, flowfl, iginfl, strpfl, xonfl;
    unsigned int baudrate;
    char altpath[256], baudstr[16], buf[4096], dev[256], filstr[384];
    char lgflstr[256], mailstr[64], mestr[512], name[64], obuf[4096];
    char path[256], pidf[384], pidir[256];
    char sendch;
    uid_t uid;
    gid_t gid;
    struct group *group;
    struct passwd *passwd;
    struct termios newtio;
    #ifdef SKLRD_USE_TIMEOUT
     struct timeval tv;
    #endif
    FILE *f1;
    fd_set readfs;

    ch     = 0;
    debfl  = 0;		/*  Debug on when 1  */
    flowfl = 0;		/*  CRTSCTS when 1  */
    gid    = 0;		/*  Group Id  */
    i      = 0;
    iginfl = 0;		/*  Ignore stdin when 1  */
    len    = 0;		/*  Length  */
    maxfd  = 0;
    op     = 0;		/*  Output pointer  */
    opt    = 0;
    pid    = 0;		/*  Process Id  */
    retcnt = 0;		/*  Retry count  */
    retval = 0;
    sendch = 0;		/*  Char to be send to Tty  */
    strpfl = 0;		/*  Strip certain data  */
    tot    = 0;		/*  Total bytes written  */
    uid    = 0;		/*  User Id  */
    xonfl  = 0;		/*  XON XOFF when 1  */
    memset(altpath, 0, sizeof(altpath));	/*  Alternative path  */
    memset(buf,     0, sizeof(buf));
    memset(datstr,  0, sizeof(datstr));		/*  Date time string  */
    memset(dev,     0, sizeof(dev));		/*  Tty  */
    memset(&dstv,   0, sizeof(dstv));
    memset(filstr,  0, sizeof(filstr));		/*  Output file name  */
    memset(lgflstr, 0, sizeof(lgflstr));	/*  Logfile name  */
    memset(logstr,  0, sizeof(logstr));		/*  String for logfile  */
    memset(mailstr, 0, sizeof(mailstr));	/*  Email address of admin  */
    memset(mestr,   0, sizeof(mestr));		/*  String for email message  */
    memset(name,    0, sizeof(name));		/*  File name prefix  */
    memset(&newtio, 0, sizeof(newtio));		/*  This makes the tty rather raw  */
    memset(obuf,    0, sizeof(obuf));		/*  Ouput buffer  */
    memset(path,    0, sizeof(path));		/*  Path to output files  */
    #ifndef SKLRD_USE_TIMEOUT
     memset(pfd,     0, sizeof(pfd));		/*  Pipe Fd  */
    #endif
    memset(pidf,    0, sizeof(pidf));		/*  Pid file  */
    memset(pidir,   0, sizeof(pidir));		/*  Pid directory  */
    memset(systr,   0, sizeof(systr));		/*  String for system()  */
    memset(tologst, 0, sizeof(tologst));
    #ifdef SKLRD_USE_TIMEOUT
     memset(&tv,     0, sizeof(tv));
    #endif
    f1     = NULL;
    group  = NULL;
    logfil = NULL;
    passwd = NULL;

    /*  Default speed  */
    strncpy(baudstr, SKLRD_DFLT_SPD, 15);
    /*  Continue program  */
    contprog = 1;
    /*  Get the pid  */
    pid = getpid();

    /*  Command line processing  */
    while ((opt = getopt(argc, argv, "a:b:cd:Dg:hil:m:n:p:r:su:vx")) != -1) {
	switch (opt) {
	    case 'a':
		strncpy(altpath, optarg, 255);
		break;
	    case 'b':
		strncpy(baudstr, optarg, 15);	/*  < 255  */
		break;
	    case 'c':
		/*  Use RTS CTS flow control  */
		flowfl = 1;
		break;
	    case 'd':
		strncpy(dev, optarg, 255);
		break;
	    case 'D':
		debfl = 1;
		break;
	    case 'g':
		gid = (gid_t) atoi(optarg);
		if (gid < 1 || gid > 65535) {
		    /*  Group not numeric or invalid  */
		    group = getgrnam(optarg);
		    if (group == NULL) {
			fprintf(stderr, "Group not found\n");
			exit(1);
		    } else {
			gid = group->gr_gid;
		    }
		}
		break;
	    case 'h':
		hlp();
		exit(0);
	    case 'i':
		/*  Ignore stdin  */
		iginfl = 1;
		break;
	    case 'l':
		strncpy(lgflstr, optarg, 255);
		break;
	    case 'm':
		strncpy(mailstr, optarg, 63);	/*  < 255  */
		break;
	    case 'n':
		strncpy(name, optarg, 63);	/*  < 255  */
		break;
	    case 'p':
		strncpy(path, optarg, 255);
		break;
	    case 'r':
		strncpy(pidir, optarg, 255);
		break;
	    case 's':
		strpfl = 1;
		break;
	    case 'u':
		uid = (uid_t) atoi(optarg);
		if (uid < 1 || uid > 65535) {
		    /*  User not numeric or invalid  */
		    passwd = getpwnam(optarg);
		    if (passwd == NULL) {
			fprintf(stderr, "User not found\n");
			exit(1);
		    } else {
			uid = passwd->pw_uid;
		    }
		}
		break;
	    case 'v':
		prtvers();
		exit(0);
	    case 'x':
		xonfl = 1;
		break;
	    default:
		hlp();
		exit(1);
	}
    }

    /*  Open logfile if specified  */
    if (lgflstr[0] != 0) {
	if((logfil = fopen(lgflstr, "a")) == NULL)
	    fprintf(stderr, "Can't open logfile %s: %s\n", 
	      lgflstr, strerror(errno));
    }

    if ((baudrate = findbaud(baudstr)) == 0) { 
	prtolog("Invalid baudrate:", baudstr);
	if (logfil == NULL) {
	    /*  List in interactive mode only  */
	    fprintf(stderr, "baudrate must be one of: ");
	    listbauds();
	    fprintf(stderr, ".\n");
	} else {
	    fclose(logfil);
	}
	exit(1);
    }
    if (dev[0] == 0) {
	prtolog("Must specify device. EG: -d /dev/ttyS1", NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  File name. If no prefix is specified, use tty name instead  */
    if (name[0] == 0) {
	/*  Strip path from tty  */
	i = strlen(dev) - 1;	/*  Start op last char  */
        while (i > 0) {
	    if (dev[i] == '/') {
		strncpy(name, dev + i + 1, 63);
		break;
	    }
	    i--;
	}
    } else {
	len = strlen(name) - 1;
	if (name[len] == '/') {
	    /*  Strip strailing slash  */
	    name[len] = 0;
	}
    }
    if (name[0] == 0) {
	prtolog("Could not set filename.", NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  Path  */
    len = strlen(path) - 1;
    if (path[len] == '/') {
	/*  Strip strailing slash  */
	path[len] = 0;
    }

    #ifndef SKLRD_USE_TIMEOUT
     /*  Pipe to assist quit on SIGTERM  */
     if (pipe(pfd) < 0) {
	snprintf(logstr, 511, "Can't open pipe: %s", strerror(errno));
	prtolog(logstr, NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
     }
    #endif

    /*
    *	Open modem device for reading and writing and not as controlling tty
    *	because we don't want to get killed if line noise sends CTRL-C.
    */
    if ((fd = open(dev, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0) {
	snprintf(logstr, 511, "Can't open device %s: %s", dev, strerror(errno));
	prtolog(logstr, NULL);
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  Pidfile name  */
    if (pidir[0] == 0) {
	snprintf(pidf, 383, "/run/%s.pid", name);
    } else {
	len = strlen(pidir) - 1;
	if (pidir[len] == '/') {
	    /*  Strip strailing slash  */
	    pidir[len] = 0;
	}
	snprintf(pidf, 383, "%s/%s.pid", pidir, name);
    }

    /*  UID and GID  */
    if (uid != 0) {
	if (gid == 0)
	    gid = uid;
	if (setgid(gid) < 0) {
	    snprintf(logstr, 511, "Can't set GID to %d: %s", \
	      gid, strerror(errno));
	    prtolog(logstr, NULL);
	    gid = 0;
	}
	if (setuid(uid) < 0) {
	    snprintf(logstr, 511, "Can't set UID to %d: %s", \
	      uid, strerror(errno));
	    prtolog(logstr, NULL);
	    uid = 0;
	}
    }

    /*  Open pid file as new user  */
    if ((f1 = fopen(pidf, "w")) == NULL) {
	snprintf(logstr, 511, "Can't open pid file %s: %s", \
	  pidf, strerror(errno));
	prtolog(logstr, NULL);
	pidf[0] = 0;
    } else {
	fprintf(f1, "%d\n", pid);
	fclose(f1);
    }
    len = 0;
    f1 = NULL;

    /*  save current serial port settings  */
    tcgetattr(fd, &oldtio);

    /*
    BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
    CRTSCTS:  Output hardware flow control.
    CLOCAL:   Ignore modem control (DSR, DCD) lines.
    CREAD:    Enable receiving characters.
    CS8:      8 bit, no parity, 1 stopbit.
    CREAD:    enable receiving characters
    HUPCL:    Lower modem control lines (DTR) when done.
    IXON:     Enable software flow control (outgoing)
    IXOFF:    Enable software flow control (incoming)
    */
    newtio.c_cflag = baudrate | HUPCL | CLOCAL | CREAD | CS8;
    if (flowfl != 0)
	newtio.c_cflag = newtio.c_cflag | CRTSCTS;
    if (xonfl  != 0)
	newtio.c_cflag = newtio.c_cflag | IXON | IXOFF;
    /*
    *	Input: IGNPAR: ignore bytes with parity errors
    */
    newtio.c_iflag = IGNPAR;
    /*
    *	Blocking read until 1 character arrives.
    *	You don't need this with a non blocking read.
    */
    /* newtio.c_cc[VMIN] = 1; */

    /*  Clean the modem line and activate the settings for the port  */
    tcflush(fd, TCIOFLUSH);
    if (tcsetattr(fd, TCSANOW, &newtio) != 0) {
	snprintf(logstr, 511, "Can't set %s attributes: %s", dev, strerror(errno));
	prtolog(logstr, NULL);
	resetty();
	if(logfil != NULL)
	    fclose(logfil);
	exit(1);
    }

    /*  Print info  */
    prtolog("Skyla-read version:", skyvers);
    #ifdef SKLRD_USE_TIMEOUT
     prtolog("Compiled with use timeout", NULL);
    #endif
    snprintf(logstr, 511, "Device %s, speed %s bps", dev, baudstr);
    prtolog(logstr, NULL);
    if(flowfl != 0)
	prtolog("RTS CTS flow control", NULL);
    if(xonfl  != 0)
	prtolog("XON XOFF flow control", NULL);
    prtolog("Prefix:", name);
    if(path[0] != 0)
	prtolog("Files saved in:", path);
    if(pidf[0] != 0)
	prtolog("Pid file:", pidf);
    if(uid != 0 || gid != 0) {
	snprintf(logstr, 511, "Uid: %d, Gid: %d", uid, gid);
	prtolog(logstr, NULL);
    }
    if (iginfl == 0 && logfil == NULL)
	printf("Press 'q' <Enter> to quit.\n");

    /*  Catch sigals  */
    //signal(SIGHUP,  stop);
    signal(SIGTERM,  stop);

    while (contprog != 0) {
	if (sendch != 0) {
	    /*  Write stuff before select  */
	    retcnt = 0;
	    if (write(fd, &sendch, 1) < 1) {
		if (errno == EAGAIN || errno == EINTR) {
		    prtolog("Retrying write to serial port", NULL);
		} else {
		    snprintf(logstr, 511, "Can't write to tty: %s", \
		      strerror(errno));
		    resetty();
		    fclose(f1);
		    if(logfil != NULL)
			fclose(logfil);
		    exit(1);
		}
		retcnt++;
		if (retcnt > 32 ) {
		    prtolog("Too many retries writing to serial port", NULL);
		    resetty();
		    fclose(f1);
		    if(logfil != NULL)
			fclose(logfil);
		    exit(1);
		}
		if (debfl != 0)
		    prtolog("Debug: Wrote 1 byte to serial port", NULL);
	    }
	    sendch = 0;	/*  Done  */
	}

	/*  XXX SIGTERM Test XXX  */
	//fprintf(stderr, "Input here\n");
	//read(0, &ch, 1);

	/*  (Re) Initialise select()  */
	FD_ZERO(&readfs);
	if (iginfl == 0) {
	    /*  Do not ignore stdin  */
	    FD_SET(0, &readfs);
	}
	FD_SET(fd, &readfs);
	#ifdef SKLRD_USE_TIMEOUT
	 /*  Set the timeout to 1 s  */
	 tv.tv_sec  = 1;
	 tv.tv_usec = 0;
	 maxfd = fd + 1;
	 retval = select(maxfd, &readfs, NULL, NULL, &tv);
	#else
	 FD_SET(pfd[0], &readfs);
	 if (pfd[0] > fd)
	     maxfd = pfd[0] + 1;
	 else
	     maxfd = fd + 1;
	 retval = select(maxfd, &readfs, NULL, NULL, NULL);
	#endif
	if (retval < 0) {
	    if(contprog == 0)
		break;
	    if (errno == EINTR) {
		prtolog("Reinitilizing select()", NULL);
		continue;
	    } else {
		snprintf(logstr, 511, "Select failed: %s", strerror(errno));
		prtolog(logstr, NULL);
		resetty();
		fclose(f1);
		if(logfil != NULL)
		    fclose(logfil);
		exit(1);
	    }
	}

	/*  Block until read  */
	if (iginfl == 0 && FD_ISSET(0, &readfs)) {
	    /*  Check continue first  */
	    if(contprog == 0)
		break;
	    /*  Read data from keyboard  */
	    read(0, &ch, 1);
	    if (ch == 'Q' || ch == 'q') {
		prtolog("Quitting", NULL);
		contprog = 0;	/*  Don't continue  */
	    }
	}
	#ifndef SKLRD_USE_TIMEOUT
	 if (FD_ISSET(pfd[0], &readfs)) {
	     if(contprog == 0)
		 break;
	     /*  Read data from pipe  */
	     read(pfd[0], &ch, 1);
	     contprog = 0;
	 }
	#endif
	if (FD_ISSET(fd, &readfs)) {
	    /*  Check continue first  */
	    if(contprog == 0)
		break;
	    /*  Read data from ttySx  */
	    len = read(fd, buf, 4095);
	    if (len < 1) {
		if(contprog == 0)
		    break;
		if (errno == EAGAIN || errno == EINTR) {
		    prtolog("Reinitilizing select()", NULL);
		    continue;
		} else {
		    snprintf(logstr, 511, "Can't read from tty: %s", \
		      strerror(errno));
		    prtolog(logstr, NULL);
		    resetty();
		    fclose(f1);
		    if(logfil != NULL)
			fclose(logfil);
		    exit(1);
		}
	    }
	    buf[len] = 0;	/*  Terminate  */
	    if (debfl != 0) {
		snprintf(logstr, 511, \
		  "Debug: Received %d bytes from serial port", len);
		prtolog(logstr, NULL);
	    }
	    i = 0;
	    sendch = 0;
	    while (i < len) {
		/*  Check comm prot per byte  */
		if (buf[i] == 0x01) {
		    /*  SOH  */
		    sendch = 0x06;	/*  ACK  */
		    if (f1 == NULL) {
			/*  Open file  */
			if(debfl != 0)
			    prtolog("Debug: Opening file", NULL);
			op = 0;
			/* FILE * genfil(FILE *f1, char *s, char *path, char *prefix) */
			f1 = genfil(f1, filstr, path, name);
			if (f1 == NULL) {
			    prtolog("Can't open file", NULL);
			    if (mailstr[0] != 0) {
				/*  Mail error to admin  */
				snprintf(mestr, 511, \
				  "Can't open file %s", filstr);
				mailerr(mestr, mailstr);
			    }
			    if (altpath[0] == 0) {
				resetty();
				if(logfil != NULL)
				    fclose(logfil);
				exit(1);
			    } else {
				f1 = genfil(f1, filstr, altpath, name);
				if (f1 == NULL) {
				    prtolog("Can't open file", NULL);
				    resetty();
				    if(logfil != NULL)
					fclose(logfil);
				    exit(1);
				}
			    }
			}
			if(debfl != 0)
			    prtolog("Debug: Opened file", filstr);
			//tot = 0;	/*  New file  */
			tot = fprintf(f1, "Saved to file at: %s\n\n", \
			  gendat(datstr, 1));
			if (debfl != 0) {
			    snprintf(logstr, 511, \
			      "Debug: Wrote %d bytes to file", tot);
			      prtolog(logstr, NULL);
			}
		    } else {
			prtolog("File should be closed", NULL);
		    }
		} else if (buf[i] == 0x04) {
		    /*  EOT  */
		    sendch = 0x06;	/*  ACK  */
		    if (f1 == NULL) {
			prtolog("File should be open", NULL);
		    } else {
			/*  Close file  */
			if(debfl != 0)
			    prtolog("Debug: Closing file", filstr);
			fflush(f1);
			fclose(f1);
			f1 = NULL;	/*  Indicate that file is closed  */
			if (debfl != 0) {
			    snprintf(logstr, 511, \
			      "Debug: Wrote %d bytes to file %s", \
			      tot, filstr);
			    prtolog(logstr, NULL);
			}
		    }
		} else {
		    /*  Write to file  */
		    if (f1 != NULL && buf[i] != 0) {
			obuf[op] = buf[i];
			if (obuf[op] == 10 || op > 4080) {
			    /*  Terminate  */
			    obuf[op + 1] = 0;
			    /*  Strip trailing whitespace  */
			    op = strptws(obuf, op);
			    /*  Skip stuff  */
			    if (strpfl == 0 || \
			      ((obuf[0] < '0' || obuf[0] > '9') && \
			      fndskip(obuf) == 0)) {
				if (fprintf(f1, "%s", obuf) < 1) {
				    prtolog("Can't write fo file", NULL);
				    resetty();
				    if(logfil != NULL)
					fclose(logfil);
				    exit(1);
				}
				/*  Op is 'pointer', so add 1  */
				tot = tot + op + 1;
				if (debfl != 0) {
				    snprintf(logstr, 511, \
				      "Debug: Wrote %d bytes to file", op + 1);
				    prtolog(logstr, NULL);
				}
			    }
			    op = 0;
			    obuf[0] = 0;
			} else {
			    op++;
			}
		    }
		}
		i++;
	    }	/*  End of while (i < len)  */
	}	/*  End of FD_ISSET(fd, &readfs)  */
    }		/*  End of while (contprog != 0)  */
    
    /*  restore the old port settings  */
    resetty();
    /*  Signals to default  */
    //signal(SIGHUP,  SIG_DFL);
    signal(SIGTERM,  SIG_DFL);
    /*  Close file if open  */
    if (f1 != NULL) {
	fflush(f1);
	fclose(f1);
    }
    #ifndef SKLRD_USE_TIMEOUT
     /*  Close pipe  */
     close(pfd[0]);
     close(pfd[1]);
    #endif

    /*    Remove PID file    */
    if (pidf[0] != 0) {
	if(unlink(pidf) != 0)
	    prtolog("Can't remove pidfile:", pidf);
    }

    prtolog("skyla-read stopped", NULL);

    if(logfil != NULL)
	fclose(logfil);

    return(0);
}

